import re

from textual.app import ComposeResult
from textual.binding import Binding
from textual.containers import Horizontal, Vertical
from textual.screen import ModalScreen
from textual.widgets import Button, Checkbox, Input, Label, Static

from dolphie.DataTypes import ConnectionSource, HotkeyCommands
from dolphie.Widgets.autocomplete import AutoComplete, Dropdown, DropdownItem


class CommandModal(ModalScreen):
    CSS = """
        CommandModal > Vertical {
            background: #131626;
            border: tall #384673;
            height: auto;
            width: auto;
        }
        CommandModal > Vertical > * {
            width: auto;
            height: auto;
            content-align: center middle;
        }
        CommandModal #filter_container {
            width: auto;
            height: auto;
        }
        CommandModal #kill_container {
            width: auto;
            height: auto;
        }
        CommandModal #filter_container Input {
            width: 60;
            border-title-color: #d2d2d2;
        }
        CommandModal #kill_container Input {
            width: 60;
            border-title-color: #d2d2d2;
        }
        CommandModal Label {
            width: 100%;
            content-align: center middle;
            padding-bottom: 1;
        }
        CommandModal #error_response {
            color: #fe5c5c;
            width: 100%;
            height: auto;
            content-align: center middle;
            padding-bottom: 1;
        }
        CommandModal Checkbox {
            background: #131626;
            border: none;
            content-align: center middle;
            padding-top: 1;
            width: 100%;
        }
        CommandModal #sleeping_queries {
            padding-bottom: 1;
        }
    """
    BINDINGS = [
        Binding("escape", "app.pop_screen", "", show=False),
    ]

    def __init__(
        self,
        command,
        message,
        connection_source: ConnectionSource = None,
        processlist_data=None,
        host_cache_data=None,
        max_replay_timestamp=None,
    ):
        super().__init__()
        self.command = command
        self.message = message
        self.connection_source = connection_source
        self.processlist_data = processlist_data
        self.host_cache_data = host_cache_data
        self.max_replay_timestamp = max_replay_timestamp

        self.dropdown_items = []
        if processlist_data:
            sorted_keys = sorted(processlist_data.keys(), key=lambda x: int(x))
            self.dropdown_items = [DropdownItem(thread_id) for thread_id in sorted_keys]

    def compose(self) -> ComposeResult:
        with Vertical():
            with Vertical():
                yield Label(f"[b]{self.message}[/b]")
                with Vertical(id="filter_container"):
                    yield AutoComplete(
                        Input(id="filter_by_username_input"), Dropdown(id="filter_by_username_dropdown_items", items=[])
                    )
                    yield AutoComplete(
                        Input(id="filter_by_host_input"), Dropdown(id="filter_by_host_dropdown_items", items=[])
                    )
                    yield AutoComplete(
                        Input(id="filter_by_db_input"), Dropdown(id="filter_by_db_dropdown_items", items=[])
                    )
                    yield AutoComplete(
                        Input(id="filter_by_hostgroup_input"),
                        Dropdown(id="filter_by_hostgroup_dropdown_items", items=[]),
                    )
                    yield Input(id="filter_by_query_time_input")
                    yield Input(id="filter_by_query_text_input")
                with Vertical(id="kill_container"):
                    yield AutoComplete(
                        Input(id="kill_by_username_input"), Dropdown(id="kill_by_username_dropdown_items", items=[])
                    )
                    yield AutoComplete(
                        Input(id="kill_by_host_input"), Dropdown(id="kill_by_host_dropdown_items", items=[])
                    )
                    yield Input(id="kill_by_age_range_input", placeholder="Example: 5-8")
                    yield Input(id="kill_by_query_text_input")
                    yield Checkbox("Include sleeping queries", id="sleeping_queries")
                    yield Label("[dark_gray][b]Note[/b]: This feature uses threads visible in the processlist")
                yield AutoComplete(
                    Input(id="modal_input"),
                    Dropdown(id="dropdown_items", items=self.dropdown_items),
                )

                yield Static(id="error_response")
            with Horizontal(classes="button_container"):
                yield Button("Submit", id="submit", variant="primary")
                yield Button("Cancel", id="cancel")

    def on_mount(self):
        input = self.query_one("#modal_input", Input)
        filter_container = self.query_one("#filter_container", Vertical)
        kill_container = self.query_one("#kill_container", Vertical)
        self.query_one("#error_response", Static).display = False

        filter_container.display = False
        kill_container.display = False

        if self.command == HotkeyCommands.thread_filter:
            input.display = False
            filter_container.display = True

            self.query_one("#filter_by_username_input", Input).focus()
            self.query_one("#filter_by_username_input", Input).border_title = "Username"
            self.query_one("#filter_by_username_dropdown_items", Dropdown).items = self.create_dropdown_items("user")
            self.query_one("#filter_by_host_input", Input).border_title = "Host/IP"
            self.query_one("#filter_by_host_dropdown_items", Dropdown).items = self.create_dropdown_items("host")
            self.query_one("#filter_by_db_input", Input).border_title = "Database"
            self.query_one("#filter_by_db_dropdown_items", Dropdown).items = self.create_dropdown_items("db")
            self.query_one("#filter_by_query_time_input", Input).border_title = "Minimum Query Time (seconds)"
            self.query_one("#filter_by_query_text_input", Input).border_title = "Partial Query Text"

            if self.connection_source != ConnectionSource.proxysql:
                self.query_one("#filter_by_hostgroup_input", Input).display = False
            else:
                self.query_one("#filter_by_host_input", Input).border_title = "Backend Host/IP"
                self.query_one("#filter_by_hostgroup_input", Input).border_title = "Hostgroup"
                self.query_one("#filter_by_hostgroup_dropdown_items", Dropdown).items = self.create_dropdown_items(
                    "hostgroup"
                )
        elif self.command == HotkeyCommands.thread_kill_by_parameter:
            input.display = False
            kill_container.display = True

            self.query_one("#kill_by_username_input", Input).focus()
            self.query_one("#kill_by_username_input", Input).border_title = "Username"
            self.query_one("#kill_by_username_dropdown_items", Dropdown).items = self.create_dropdown_items("user")
            self.query_one("#kill_by_host_input", Input).border_title = "Host/IP"
            self.query_one("#kill_by_host_dropdown_items", Dropdown).items = self.create_dropdown_items("host")
            self.query_one("#kill_by_age_range_input", Input).border_title = "Age Range (seconds)"
            self.query_one("#kill_by_query_text_input", Input).border_title = "Partial Query Text"

            sleeping_queries_checkbox = self.query_one("#sleeping_queries", Checkbox)
            sleeping_queries_checkbox.toggle()

            input.placeholder = "Select an option from above"
        elif self.command == HotkeyCommands.rename_tab:
            input.placeholder = "Colors can be added by wrapping them in []"
            input.styles.width = 50
            input.focus()
        elif self.command == HotkeyCommands.variable_search:
            input.placeholder = "Input 'all' to show everything"
            input.focus()
        elif self.command in [HotkeyCommands.show_thread, HotkeyCommands.thread_kill_by_id]:
            input.placeholder = "Input a Thread ID"
            input.focus()
        elif self.command == HotkeyCommands.refresh_interval:
            input.placeholder = "Input a refresh interval"
            input.focus()
        elif self.command == HotkeyCommands.replay_seek:
            if self.max_replay_timestamp:
                input.value = self.max_replay_timestamp
            input.placeholder = "Format: 2024-07-25 13:00:00"
            input.focus()
        else:
            input.focus()

    def create_dropdown_items(self, field):
        dropdown_items = []

        if field:
            # Filter out None values before sorting
            sorted_array = sorted(
                set(
                    getattr(thread, field)
                    for thread in self.processlist_data.values()
                    if getattr(thread, field) is not None
                )
            )
            dropdown_items = [DropdownItem(str(value)) for value in sorted_array]

        return dropdown_items

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id != "submit":
            self.app.pop_screen()
            return

        modal_input = self.query_one("#modal_input", Input).value
        if not modal_input and self.command not in [
            HotkeyCommands.rename_tab,
            HotkeyCommands.thread_kill_by_parameter,
            HotkeyCommands.thread_filter,
        ]:
            self.update_error_response("Input cannot be empty")
            return

        if self.command == HotkeyCommands.thread_filter:
            # Fetch all filter values
            filters = {
                "username": self.query_one("#filter_by_username_input", Input).value,
                "host": self.query_one("#filter_by_host_input", Input).value,
                "db": self.query_one("#filter_by_db_input", Input).value,
                "hostgroup": self.query_one("#filter_by_hostgroup_input", Input).value,
                "query_time": self.query_one("#filter_by_query_time_input", Input).value,
                "query_text": self.query_one("#filter_by_query_text_input", Input).value,
            }

            # Use IP address instead of hostname since that's what is used in the processlist
            if filters["host"]:
                filters["host"] = next(
                    (ip for ip, addr in self.host_cache_data.items() if filters["host"] == addr), filters["host"]
                )

            # Validate numeric fields
            for value, field_name in [(filters["query_time"], "Query time"), (filters["hostgroup"], "Hostgroup")]:
                if value and not re.search(r"^\d+$", value):
                    self.update_error_response(f"{field_name} must be an integer")
                    return

            # Ensure at least one filter is provided
            if not any(filters.values()):
                self.update_error_response("At least one field must be provided")
                return

            # Dismiss with the filter values
            self.dismiss(list(filters.values()))
        elif self.command == HotkeyCommands.thread_kill_by_parameter:
            # Get input values
            kill_by_username = self.query_one("#kill_by_username_input", Input).value
            kill_by_host = self.query_one("#kill_by_host_input", Input).value
            kill_by_age_range = self.query_one("#kill_by_age_range_input", Input).value
            kill_by_query_text = self.query_one("#kill_by_query_text_input", Input).value
            checkbox_sleeping_queries = self.query_one("#sleeping_queries", Checkbox).value

            age_range_lower_limit, age_range_upper_limit = None, None

            # Process and validate age range input
            if kill_by_age_range:
                match = re.match(r"(\d+)-(\d+)", kill_by_age_range)
                if match:
                    age_range_lower_limit, age_range_upper_limit = map(int, match.groups())
                    if age_range_lower_limit > age_range_upper_limit:
                        self.update_error_response("Invalid age range! Lower limit can't be higher than upper")
                        return
                else:
                    self.update_error_response("Invalid age range")
                    return

            # Ensure at least one parameter is provided
            if not any([kill_by_username, kill_by_host, kill_by_age_range, kill_by_query_text]):
                self.update_error_response("At least one parameter must be provided")
                return

            # Dismiss with the filter values
            self.dismiss(
                [
                    kill_by_username,
                    kill_by_host,
                    kill_by_age_range,
                    age_range_lower_limit,
                    age_range_upper_limit,
                    kill_by_query_text,
                    checkbox_sleeping_queries,
                ]
            )

        elif self.command in [HotkeyCommands.thread_kill_by_id, HotkeyCommands.show_thread]:
            value = next((thread_id for thread_id in self.processlist_data.keys() if modal_input == thread_id), None)

            if not value:
                self.update_error_response(f"Thread ID [bold red]{modal_input}[/bold red] does not exist")
            else:
                self.dismiss(modal_input)
        elif self.command == HotkeyCommands.refresh_interval:
            try:
                modal_input = float(modal_input)
            except ValueError:
                self.update_error_response("Input must be a number")
                return

            if modal_input <= 0:
                self.update_error_response("Input must be greater than 0")
                return

            self.dismiss(modal_input)
        elif self.command == HotkeyCommands.replay_seek:
            if not re.search(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}", modal_input):
                self.update_error_response("Invalid timestamp format")
                return

            self.dismiss(modal_input)
        else:
            self.dismiss(modal_input)

    def update_error_response(self, message):
        error_response = self.query_one("#error_response", Static)
        error_response.display = True
        error_response.update(message)

    def on_input_submitted(self):
        if self.command not in [HotkeyCommands.thread_filter, HotkeyCommands.thread_kill_by_parameter]:
            self.query_one("#submit", Button).press()



import os
import sqlite3
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Union

import orjson
import zstandard as zstd
from loguru import logger

from dolphie.DataTypes import (
    ConnectionSource,
    ProcesslistThread,
    ProxySQLProcesslistThread,
)
from dolphie.Dolphie import Dolphie
from dolphie.Modules import MetricManager
from dolphie.Modules.Functions import format_bytes, minify_query


@dataclass
class MySQLReplayData:
    timestamp: str
    global_status: dict
    global_variables: dict
    binlog_status: dict
    innodb_metrics: dict
    replica_manager: dict
    replication_status: dict
    processlist: dict
    metric_manager: dict
    metadata_locks: dict


@dataclass
class ProxySQLReplayData:
    timestamp: str
    global_status: dict
    global_variables: dict
    command_stats: dict
    hostgroup_summary: dict
    processlist: dict
    metric_manager: dict


class ReplayManager:
    """
    ReplayManager class for capturing and replaying Dolphie instance states.
    """

    def __init__(self, dolphie: Dolphie):
        """
        Initializes the ReplayManager with Dolphie instance and SQLite database settings.

        Args:
            dolphie: The Dolphie instance.
        """
        self.dolphie = dolphie
        # We will increment this to force a new replay file if the schema changes in future versions
        self.schema_version = 1
        self.connection: sqlite3.Connection = None
        self.cursor: sqlite3.Cursor = None
        self.current_index = 0  # This is used to keep track of the last primary key read from the database
        self.min_timestamp = None
        self.max_timestamp = None
        self.min_id = None
        self.max_id = None
        self.last_purge_time = datetime.now() - timedelta(hours=1)  # Initialize to an hour ago
        self.replay_file_size = 0
        self.compression_dict = None
        self.dict_samples = []

        # Determine filename used for replay file
        hostname = f"{dolphie.host}_{dolphie.port}"
        if dolphie.replay_file:
            self.replay_file = dolphie.replay_file
        elif dolphie.daemon_mode:
            self.replay_file = f"{dolphie.replay_dir}/{hostname}/daemon.db"
        elif dolphie.record_for_replay:
            self.replay_file = f"{dolphie.replay_dir}/{hostname}/{datetime.now().strftime('%Y_%m_%d_%H_%M_%S')}.db"
            dolphie.app.notify(f"File: [highlight]{self.replay_file}[/highlight]", title="Recording data", timeout=10)
        else:
            # No options specified for replaying, skip initialization
            return

        os.makedirs(os.path.dirname(self.replay_file), mode=0o770, exist_ok=True)
        logger.info(f"Replay SQLite file: {self.replay_file} ({self.dolphie.replay_retention_hours} hours retention)")

        self._initialize_sqlite()
        self._manage_metadata()

        if dolphie.replay_file:
            self._get_replay_file_metadata()

    def _initialize_sqlite(self):
        """
        Initializes the SQLite database and creates the necessary tables.
        """
        database_exists = True if os.path.exists(self.replay_file) else False

        self.connection = sqlite3.connect(self.replay_file, isolation_level=None, check_same_thread=False)
        self.cursor = self.connection.cursor()

        # Lock down the permissions of the replay file
        os.chmod(self.replay_file, 0o660)

        if not database_exists:
            logger.info("Created new SQLite database and connected to it")
        else:
            logger.info("Connected to SQLite")

        self.cursor.execute(
            """
            CREATE TABLE IF NOT EXISTS replay_data (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                timestamp DATETIME,
                data BLOB
            )"""
        )
        self.cursor.execute("CREATE INDEX IF NOT EXISTS idx_replay_data_timestamp ON replay_data (timestamp)")
        metadata_table_query = """
            CREATE TABLE IF NOT EXISTS metadata (
                schema_version INTEGER DEFAULT 1,
                hostname VARCHAR(255),
                host_version VARCHAR(255),
                host_distro VARCHAR(255),
                connection_source VARCHAR(255),
                dolphie_version VARCHAR(255),
                compression_dict BLOB
            )"""
        self.cursor.execute(metadata_table_query)

        # Enable auto-vacuum if it's not already enabled. This will help keep the database file size down.
        if self.cursor.execute("PRAGMA auto_vacuum").fetchone()[0] != 1:
            self.cursor.execute("PRAGMA auto_vacuum = FULL")
            self.cursor.execute("VACUUM")

        # This is temporary for backwards compatibility with older replay files
        columns = [info[1] for info in self.cursor.execute("PRAGMA table_info(metadata)").fetchall()]
        if "schema_version" not in columns:
            logger.warning("The 'schema_version' column is missing in the metadata table. Adding it...")

            self.cursor.execute("ALTER TABLE metadata RENAME TO metadata_old")
            self.cursor.execute(metadata_table_query)
            self.cursor.execute(
                f"""
                INSERT INTO metadata (
                    schema_version, hostname, host_version, host_distro, connection_source,
                    dolphie_version, compression_dict
                )
                SELECT
                    {self.schema_version}, hostname, host_version, host_distro, connection_source,
                    dolphie_version, compression_dict
                FROM
                    metadata_old
                """
            )
            self.cursor.execute("DROP TABLE metadata_old")

            logger.success("Column was added")

        self.purge_old_data()

    def purge_old_data(self):
        """
        Purges data older than the retention period specified by hours_of_retention.
        Only runs if at least an hour has passed since the last purge.
        """
        if not self.dolphie.record_for_replay:
            return

        current_time = datetime.now()
        if (current_time - self.last_purge_time) < timedelta(hours=1):
            return  # Skip purging if less than an hour has passed

        retention_date = (current_time - timedelta(hours=self.dolphie.replay_retention_hours)).strftime(
            "%Y-%m-%d %H:%M:%S"
        )

        self.cursor.execute("DELETE FROM replay_data WHERE timestamp < ?", (retention_date,))
        self.last_purge_time = current_time

    def seek_to_timestamp(self, timestamp: str):
        """
        Seeks to the specified timestamp in the SQLite database.

        Args:
            timestamp: The timestamp to seek to.
        """
        row = self.cursor.execute("SELECT id FROM replay_data WHERE timestamp = ?", (timestamp,)).fetchone()
        if row:
            # We subtract 1 because get_next_refresh_interval naturally increments the index
            self.current_index = row[0] - 1
            self.dolphie.app.notify(
                f"Seeking to timestamp [light_blue]{timestamp}[/light_blue]", severity="success", timeout=10
            )

            return True
        else:
            # Try to find a timestamp before the specified timestamp
            row = self.cursor.execute(
                "SELECT id, timestamp FROM replay_data WHERE timestamp < ? ORDER BY timestamp DESC LIMIT 1",
                (timestamp,),
            ).fetchone()
            if row:
                # We subtract 1 because get_next_refresh_interval naturally increments the index
                self.current_index = row[0] - 1
                self.dolphie.app.notify(
                    f"Timestamp not found, seeking to closest timestamp [light_blue]{row[1]}[/light_blue]",
                    timeout=10,
                )

                return True
            else:
                self.dolphie.app.notify(
                    f"No timestamps found on or before [light_blue]{timestamp}[/light_blue]",
                    severity="error",
                    timeout=10,
                )
                return False

    def _create_new_replay_file(self, new_replay_file: str):
        logger.info(f"Renaming replay file to: {new_replay_file}")

        os.rename(self.replay_file, new_replay_file)

        # Reset compression dict if it's already been set or else the replay file will be corrupted
        self.compression_dict = None

        self._initialize_sqlite()
        self._manage_metadata()

    def _manage_metadata(self):
        """
        Manages the metadata table with information we care about.
        """
        if not self.dolphie.record_for_replay:
            return

        row = self.cursor.execute("SELECT * FROM metadata").fetchone()
        if row is None:
            self.cursor.execute(
                "INSERT INTO metadata (schema_version, hostname, host_version, host_distro, "
                "connection_source, dolphie_version) "
                "VALUES (?, ?, ?, ?, ?, ?)",
                (
                    self.schema_version,
                    self.dolphie.host_with_port,
                    self.dolphie.host_version,
                    self.dolphie.host_distro,
                    self.dolphie.connection_source,
                    self.dolphie.app_version,
                ),
            )
        else:
            schema_version = row[0]
            host = row[1]
            host_version = row[2]
            host_distro = row[3]
            connection_source = row[4]
            app_version = row[5]
            compress_dict = row[6]

            logger.info(
                f"Replay database metadata - Host: {host}, Version: {host_version} ({host_distro}), "
                f"Dolphie: {app_version}"
            )

            if compress_dict:
                self.compression_dict = zstd.ZstdCompressionDict(compress_dict)
                logger.info(
                    f"ZSTD compression dictionary loaded (size: {format_bytes(len(compress_dict), color=False)})"
                )

            if self.dolphie.daemon_mode:
                if schema_version != self.schema_version:
                    new_replay_file = f"{self.replay_file}_v{app_version}"
                    logger.warning(
                        f"The schema version of the replay file ({schema_version}) differs from this version "
                        f"of Dolphie's schema version ({self.schema_version}). To avoid potential issues, the "
                        f"replay file will be renamed and a new one will be created"
                    )

                    self._create_new_replay_file(new_replay_file)

                # Avoid mixing connection sources in the same replay file
                if connection_source != self.dolphie.connection_source:
                    logger.critical(
                        f"The connection source of the daemon's replay file ({connection_source}) "
                        f"differs from the current connection source ({self.dolphie.connection_source}). "
                        "You should never mix connection sources in the same replay file. Please rename "
                        "the daemon's replay file and restart the daemon."
                    )

    def _get_replay_file_metadata(self):
        """
        Retrieves the replay's metadata from the metadata table.
        """
        row = self.cursor.execute("SELECT * FROM metadata").fetchone()
        if row:
            self.dolphie.host_with_port = row[1]
            self.dolphie.host_version = row[2]
            self.dolphie.host_distro = row[3]
            self.dolphie.connection_source = row[4]

            if row[6]:
                self.compression_dict = zstd.ZstdCompressionDict(row[6])
        else:
            raise Exception("Metadata not found in replay file.")

    def _train_compression_dict(self) -> bytes:
        """
        Creates a compression dictionary based on sample data to help with better compression.

        Returns:
            bytes: The created compression dictionary.
        """
        compression_dict = zstd.train_dictionary(10485760, self.dict_samples, level=5)

        logger.info(
            f"ZSTD compression dictionary trained with {len(self.dict_samples)} samples "
            f"(size: {format_bytes(len(compression_dict), color=False)})"
        )

        # Store the compression dictionary in the metadata table to be used with decompression
        self.cursor.execute("UPDATE metadata SET compression_dict = ?", (compression_dict.as_bytes(),))

        return compression_dict

    def _compress_data(self, data: str) -> bytes:
        """
        Compresses data using zstd with an optional compression dictionary.

        Args:
            data (str): Data to compress.
            dict_bytes (bytes, optional): The compression dictionary.

        Returns:
            bytes: Compressed data.
        """

        compressor = zstd.ZstdCompressor(level=5, dict_data=self.compression_dict)

        return compressor.compress(data)

    def _decompress_data(self, compressed_data: bytes) -> bytes:
        """
        Decompresses data using zstd.

        Args:
            compressed_data (bytes): Compressed data.

        Returns:
            str: Decompressed data.
        """

        decompressor = zstd.ZstdDecompressor(dict_data=self.compression_dict)

        return decompressor.decompress(compressed_data)

    def _condition_metrics(self, metric_manager: MetricManager.MetricManager):
        """
        Captures the metrics from the metric manager and returns them in a structured format.

        Args:
            metric_manager: The metric manager to capture metrics from.

        Returns:
            dict: A dictionary of captured metrics.
        """
        metrics = {"datetimes": metric_manager.datetimes}

        for metric_instance_name, metric_instance_data in metric_manager.metrics.__dict__.items():
            # Skip if the metric instance is not for the current connection source
            if self.dolphie.connection_source == ConnectionSource.mysql:
                if (
                    ConnectionSource.mysql not in metric_instance_data.connection_source
                    or not metric_instance_data.use_with_replay
                ):
                    continue
            elif (
                self.dolphie.connection_source == ConnectionSource.proxysql
                and ConnectionSource.proxysql not in metric_instance_data.connection_source
            ):
                continue

            metric_entry = metrics.setdefault(metric_instance_name, {})
            for k, v in metric_instance_data.__dict__.items():
                if not hasattr(v, "values") or not v.values:
                    continue

                metric_entry.setdefault(k, [])
                if v.values:
                    metric_entry[k] = v.values

        return metrics

    def capture_state(self):
        """
        Captures the current state of the Dolphie instance and stores it in the SQLite database.
        """
        if not self.dolphie.record_for_replay:
            return

        # Convert the dictionary of processlist objecs to a list of dictionaries with JSON
        processlist = [
            {**thread_data, "query": minify_query(thread_data["query"])} if "query" in thread_data else thread_data
            for thread_data in (v.thread_data for v in self.dolphie.processlist_threads.values())
        ]

        timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        # Prepare data dictionary
        data_dict = {
            "global_status": self.dolphie.global_status,
            "global_variables": self.dolphie.global_variables,
            "processlist": processlist,
            "metric_manager": self._condition_metrics(self.dolphie.metric_manager),
        }

        if self.dolphie.connection_source == ConnectionSource.mysql:
            # Remove some global status variables that are not useful for replaying
            keys_to_remove = [
                var_name
                for var_name in self.dolphie.global_status.keys()
                if any(
                    exclude_var in var_name.lower()
                    for exclude_var in ["performance_schema", "mysqlx", "ssl", "rsa", "tls"]
                )
            ]

            for key in keys_to_remove:
                self.dolphie.global_status.pop(key)

            # Add MySQL specific data to the dictionary
            data_dict.update(
                {
                    "binlog_status": self.dolphie.binlog_status,
                    "innodb_metrics": self.dolphie.innodb_metrics,
                    "replica_manager": self.dolphie.replica_manager.available_replicas,
                    "replication_status": self.dolphie.replication_status,
                    "metadata_locks": self.dolphie.metadata_locks,
                }
            )
        else:
            # Add ProxySQL specific data to the dictionary
            data_dict.update(
                {
                    "command_stats": self.dolphie.proxysql_command_stats,
                    "hostgroup_summary": self.dolphie.proxysql_hostgroup_summary,
                }
            )

        data_dict_bytes = orjson.dumps(data_dict)

        # Store the first 10 samples to create a compression dictionary via training
        if not self.compression_dict:
            if len(self.dict_samples) < 10:
                self.dict_samples.append(data_dict_bytes)
            else:
                self.compression_dict = self._train_compression_dict()

                # Remove the samples to save memory
                del self.dict_samples

        # Execute the SQL insert using the constructed dictionary
        self.cursor.execute(
            "INSERT INTO replay_data (timestamp, data) VALUES (?, ?)",
            (
                timestamp,
                self._compress_data(data_dict_bytes),
            ),
        )

        self.purge_old_data()

        if not self.dolphie.daemon_mode:
            self.replay_file_size = os.path.getsize(self.replay_file)

    def get_next_refresh_interval(self) -> Union[MySQLReplayData, ProxySQLReplayData, None]:
        """
        Gets the next refresh interval's data from the SQLite database and returns it as a ReplayData object.

        Returns:
            ReplayData: The next replay data.
        """

        # Get the min and max timestamps and IDs from the database so we can update the UI
        row = self.cursor.execute("SELECT MIN(timestamp), MAX(timestamp), MIN(id), MAX(id) FROM replay_data").fetchone()
        if row:
            self.min_timestamp = row[0]
            self.max_timestamp = row[1]
            self.min_id = row[2]
            self.max_id = row[3]

        # Get the next row
        row = self.cursor.execute(
            "SELECT id, timestamp, data FROM replay_data WHERE id > ? ORDER BY id LIMIT 1",
            (self.current_index,),
        ).fetchone()
        if not row:
            return None

        self.current_index = row[0]

        # Decompress and parse the JSON data
        data = orjson.loads(self._decompress_data(row[2]))

        processlist = {}
        common_params = {
            "timestamp": row[1],
            "global_status": data.get("global_status", {}),
            "global_variables": data.get("global_variables", {}),
            "metric_manager": data.get("metric_manager", {}),
        }
        if self.dolphie.connection_source == ConnectionSource.mysql:
            # Re-create the ProcesslistThread object for each thread in the JSON's processlist
            for thread_data in data["processlist"]:
                processlist[str(thread_data["id"])] = ProcesslistThread(thread_data)

            return MySQLReplayData(
                **common_params,
                binlog_status=data.get("binlog_status", {}),
                innodb_metrics=data.get("innodb_metrics", {}),
                replica_manager=data.get("replica_manager", {}),
                replication_status=data.get("replication_status", {}),
                metadata_locks=data.get("metadata_locks", {}),
                processlist=processlist,
            )
        elif self.dolphie.connection_source == ConnectionSource.proxysql:
            # Re-create the ProxySQLProcesslistThread object for each thread in the JSON's processlist
            for thread_data in data["processlist"]:
                processlist[str(thread_data["id"])] = ProxySQLProcesslistThread(thread_data)

            return ProxySQLReplayData(
                **common_params,
                command_stats=data.get("command_stats", {}),
                hostgroup_summary=data.get("hostgroup_summary", {}),
                processlist=processlist,
            )
        else:
            self.dolphie.app.notify("Invalid connection source for replay data", severity="error")
